import re
import SEP.par
import SEP.superset
def extract_pars(pars,prefix):
  out=SEP.par.sep_par(name=prefix)
  srS=re.compile("%s(\S+)"%prefix)
  for key,val in pars.return_dict().items():
    b=srS.search(key)
    if b: out.add_param(b.group(1),val)
  return out

def add_pars(par_big,par_small,prefix):
  for key,val in par_small.return_dict().items():
    par_big.add_param("%s%s"%(prefix,key),val)
  return par_big
  


class tomo_obj:
  def __init__(self,name,**kw):
    self.name=name
    self.set_defaults(kw)
    self.kw=self.update_kw(kw)

  def run_com(self,**kw):
    SEP.util.err("Must override run_com")

  def set_defaults(self,kw):
    self.kw={}

  def update_kw(self,kw):
    for key,val in kw.items(): 
      self.kw[key]=val
    return self.kw

  def par(self,r,default=None,err=None,kuse=None):
    if kuse: k=kuse
    else: k=self.kw
    if k.has_key(r): return k[r]
    elif default: 
      k[r]=default
      return default
    elif err: SEP.util.err("Missing required parameter %s in %s"%(r,self.name))
    else: return default
 
        
 
    
    

class mig_obj(tomo_obj):
  def __init__(self,name,**kw):
    tomo_obj.__init__(self,name,**kw)
  def run_com(self,**kw):
    krun=self.update_kw(kw)
    clean=self.par("clean",0)
    verb=self.par("verb",0)
    ident=self.par("ident",err=1,kuse=krun)
    vel=self.par("vel_in",err=1,kuse=krun)
    krun["image_zero"]="image_zero.%s.H"%ident
    krun["migration"]="mig.%s.H"%ident
    com="migrate.py image_zero=%s D=%s migration=%s %s ident=%s"%(self.par("image_zero",
     kuse=krun), self.par("freq",err=1,kuse=krun),
     self.par("migration",kuse=krun),self.par("mig_pars",err=1,kuse=krun).return_par_string(),
     ident)
    com=com+" clean=%d verb=%d vel_in=%s"%(clean,verb,vel)
    if not clean: 
      krun["image"]="image.%s.H"%ident
      com=com+" image=%s"%self.par("image",err=1)
    return krun,[com]
  def set_defaults(self,kw):
    self.kw={}
    self.kw["mig_pars"]=SEP.par.sep_pars(name="mig_pars")
    self.kw["mig_pars"].add_param("nref",4)
    self.kw["mig_pars"].add_param("velocity","y")
    self.kw["mig_pars"].add_param("image_real","y")

class gamma_obj(tomo_obj):
  def __init__(self,name,**kw):
    tomo_obj.__init__(self,name,**kw)
  def run_com(self,**kw):
    krun=self.update_kw(kw)
    ident=self.par("ident",err=1,kuse=krun)
    clean=self.par("clean",0,kuse=krun)
    verb=self.par("verb",0,kuse=krun)
    pars=SEP.par.sep_pars(name="gamma_pars")
    add_pars(pars,self.par("dip_pars",err=1,kuse=krun),"dip_")
    pars.add_param("ident",ident)
    mask=self.par("semb_mask")
    if mask: pars.add_param("mask",mask)
    krun["gamma"]="gamma.%s.H"%ident
    krun["var"]="var.%s.H"%ident
    pars.add_param("verb",verb)
    pars.add_param("var",self.par("var",err=1,kuse=krun))
    pars.add_param("gamma",self.par("gamma",err=1,kuse=krun))
    pars.add_param("migration",self.par("migration",err=1,kuse=krun))
    pars.add_param("image0",self.par("image_zero",err=1,kuse=krun))
    add_pars(pars,self.par("semb_pars",err=1,kuse=krun),"semb_")
    add_pars(pars,self.par("reduce_pars",err=1,kuse=krun),"reduce_")
    add_pars(pars,self.par("pick_pars",err=1,kuse=krun),"pick_")
    pars.add_param("clean",clean)
    return krun,["gamma.py %s"%pars.return_par_string()]
    
  def set_defaults(self,kw):
    self.kw={}
    self.kw["dip_pars"]=SEP.par.sep_pars(name="dip_pars")
    self.kw["semb_pars"]=SEP.par.sep_pars(name="semb_pars")
    self.kw["semb_pars"].add_string("nsmooth=12 ogam=.7 dgam=.005 ngam=121")
    self.kw["reduce_pars"]=SEP.par.sep_pars(name="reduce_pars")
    v=SEP.superset.superset(file=self.par("vel_in",err=1,kuse=kw))
    self.kw["reduce_pars"].add_string("rect1=8 rect2=5 rect3=5")
    n1,o1,d1,label,unit=v.axis(1)
    n2,o2,d2,label,unit=v.axis(2)
    n3,o3,d3,label,unit=v.axis(3)
    if d1>1: 
      s1=30.
      s2=60.
      s3=60
    else:
      s1=.03
      s2=.06
      s3=.06
    n1=o1+(d1*(n1-1))/s1
    n2=o2+(d2*(n2-1))/s2
    print "what jeaht th",n3
    n3=o3+(d3*(n3-1))/s3
    if n3==0: n3=1
    self.kw["reduce_pars"].add_string("ns=%d,%d,%d"%(n1,n2,n3))
    self.kw["reduce_pars"].add_string("os=%f,%f,%f"%(o1,o2,o3))
    self.kw["reduce_pars"].add_string("ds=%f,%f,%f"%(s1,s2,s3))
    self.kw["pick_pars"]=SEP.par.sep_pars(name="pick_pars")
    self.kw["pick_pars"].add_string("range_min=.08 range_max=.16 eps_min=.1 eps_max=,2")
    self.kw["pick_pars"].add_string("nl_iter=9 pow=1.")
    

class back_obj(tomo_obj):
  def __init__(self,name,**kw):
    tomo_obj.__init__(self,name,**kw)
  def run_com(self,**kw):
    krun=self.update_kw(kw)
    ident=self.par("ident",err=1,kuse=krun)
    verb=self.par("verb",0,kuse=krun)
    clean=self.par("clean",0,kuse=krun)
    pars=SEP.par.sep_pars(name="tomo_pars")
    pars.add_param("ident",ident)
    pars.add_param("verb",verb)
    pars.add_param("vel_in",self.par("vel_in",err=1))
    pars.add_param("gamma",self.par("gamma",err=1))
    pars.add_param("var",self.par("var",err=1))
    pars.add_param("vel_out",self.par("vel_out",err=1))
    pars.add_param("mig",self.par("migration",err=1))
    pars.add_param("image",self.par("image_zero",err=1))
    add_pars(pars,self.par("tomo_space_pars",err=1,kuse=krun),"tomo_space_")
    add_pars(pars,self.par("tomo_pars",err=1,kuse=krun),"tomo_")
    pars.add_param("clean",clean)
    return krun,["back.py %s"%pars.return_par_string()];

  def set_defaults(self,kw):
    self.kw={}
    self.kw["tomo_space_pars"]=SEP.par.sep_pars(name="tomo_space_pars")
    v=SEP.superset.superset(file=self.par("vel_in",err=1,kuse=kw))
    n1,o1,d1,label,unit=v.axis(1)
    n2,o2,d2,label,unit=v.axis(2)
    n3,o3,d3,label,unit=v.axis(3)
    self.kw["tomo_space_pars"].add_string("n1out=%d o1out=%f d1out=%f"%(int(n1)/2,float(o1),
     float(d1)*2.))
    self.kw["tomo_space_pars"].add_string("n2out=%d o2out=%f d2out=%f"%(int(int(n2)*.7),
      float(o2)-float(d2)*.2*int(n2),float(d2)*2.))
    if n3!=1:
      self.kw["tomo_space_pars"].add_string("n3out=%d o3out=%f d3out=%f"%(int(int(n3)*.7),
        float(o3)-float(d3)*.2*int(n3),float(d3)*2.))
    self.kw["tomo_pars"]=SEP.par.sep_pars(name="tomo_pars")
    d=SEP.superset.superset(file=self.par("freq",err=1,kuse=kw))
    n,o,d,label,unit=d.axis(3)
    self.kw["tomo_pars"].add_param("max_off",(int(n)-1)*float(d))
    self.kw["tomo_pars"].add_string("verb=n tau=n first_image_pct=99 last_image_pct=32")
    self.kw["tomo_pars"].add_string("first_var_pct=75 last_var_pct=100 last_dip_coh_pct=15")
    self.kw["tomo_pars"].add_string("nloop_pts=10 max_pt_dist=5,3,3 ")
    self.kw["tomo_pars"].add_string("niter_prec=90 eps_0=.3 dip_range=70 max_mem=400")
    self.kw["tomo_pars"].add_string("steer_integrate=y model_covar_method='steer'")
     
    

class tomo_proj:
  def __init__(self,objs,**kw):
    self.kw=kw
    self.objs=objs
    
  def iter_com(self,from_string,to_string,**kw):
    kout=self.update_kw(kw)
    coms=[]
    kout["ident"]=from_string
    kout["vel_in"]="vel.%s.H"%from_string
    kout["vel_out"]="vel.%s.H"%to_string
    for obj in self.objs: 
      kout,com=obj.run_com(**kout)
      print "%s\n"%com
      coms.extend(com)
    return coms 

  def update_kw(self,kin):
    b= self.kw
    for key,val in kin.items(): b[key]=val
    return b



